#include "localsolver.h"
#include <fstream>
#include <iostream>
#include <vector>

using namespace localsolver;
using namespace std;

#define COLOR_HIGH_LOW (0)
#define HIGH_LOW_COLOR (1)
#define HIGH_COLOR_LOW (2)
#define COLOR_HIGH (3)
#define HIGH_COLOR (4)

class CarSequencingColor {
public:
    // Number of vehicles
    int nbPositions;

    // Number of options
    int nbOptions;

    // Number of classes
    int nbClasses;

    // Paint batch limit
    int paintBatchLimit;

    // Objective order
    int objectiveOrder;

    // Start position
    int startPosition;

    // Options properties
    vector<int> maxCarsPerWindow;
    vector<int> windowSize;
    vector<bool> isPriorityOption;
    bool hasLowPriorityOptions;

    // Classes properties
    vector<int> colorClass;
    vector<int> nbCars;
    vector<vector<bool>> optionsData;

    // Initial sequence
    vector<int> initialSequence;

    // LocalSolver
    LocalSolver localsolver;

    // LS Program variable
    LSExpression sequence;

    // Objectives
    LSExpression objectiveColor;
    LSExpression objectiveHighPriority;
    LSExpression objectiveLowPriority;

    /* Read instance data */
    void readInstance(const string& fileName) {
        ifstream infile;
        infile.exceptions(ifstream::failbit | ifstream::badbit);
        infile.open(fileName.c_str());

        infile >> nbPositions;
        infile >> nbOptions;
        infile >> nbClasses;
        infile >> paintBatchLimit;
        infile >> objectiveOrder;
        infile >> startPosition;

        maxCarsPerWindow.resize(nbOptions);
        windowSize.resize(nbOptions);
        isPriorityOption.resize(nbOptions);
        hasLowPriorityOptions = false;
        for (int o = 0; o < nbOptions; ++o) {
            infile >> maxCarsPerWindow[o];
            infile >> windowSize[o];
            int tmp;
            infile >> tmp;
            isPriorityOption[o] = (tmp == 1);
            if (!isPriorityOption[o])
                hasLowPriorityOptions = true;
        }

        if (!hasLowPriorityOptions) {
            if (objectiveOrder == COLOR_HIGH_LOW)
                objectiveOrder = COLOR_HIGH;
            else if (objectiveOrder == HIGH_COLOR_LOW)
                objectiveOrder = HIGH_COLOR;
            else if (objectiveOrder == HIGH_LOW_COLOR)
                objectiveOrder = HIGH_COLOR;
        }

        optionsData.resize(nbClasses);
        nbCars.resize(nbClasses);
        colorClass.resize(nbClasses);
        for (int c = 0; c < nbClasses; ++c) {
            infile >> colorClass[c];
            infile >> nbCars[c];
            optionsData[c].resize(nbOptions);
            for (int o = 0; o < nbOptions; ++o) {
                int v;
                infile >> v;
                optionsData[c][o] = (v == 1);
            }
        }

        initialSequence.resize(nbPositions);
        for (int p = 0; p < nbPositions; ++p)
            infile >> initialSequence[p];
    }

    void solve(int limit) {
        // Declare the optimization model
        LSModel model = localsolver.getModel();

        // sequence[i] = j if class initially planned on position j is produced on position i
        sequence = model.listVar(nbPositions);

        // sequence is a permutation of the initial production plan, all indexes must appear exactly once
        model.addConstraint(model.partition(sequence));

        // Past classes (before startPosition) can not move
        for (int p = 0; p < startPosition; ++p)
            model.addConstraint(sequence[p] == p);

        // Create LocalSolver arrays to be able to access them with "at" operators
        LSExpression initials = model.array(initialSequence.begin(), initialSequence.end());
        LSExpression colors = model.array(colorClass.begin(), colorClass.end());
        LSExpression options = model.array();
        for (int c = 0; c < nbClasses; ++c) {
            LSExpression classOptions = model.array(optionsData[c].begin(), optionsData[c].end());
            options.addOperand(classOptions);
        }

        // Number of cars with option o in each window
        vector<vector<LSExpression>> nbCarsWindows;
        nbCarsWindows.resize(nbOptions);
        for (int o = 0; o < nbOptions; ++o) {
            nbCarsWindows[o].resize(nbPositions);
            for (int j = startPosition - windowSize[o] + 1; j < nbPositions; ++j) {
                nbCarsWindows[o][j] = model.sum();
                for (int k = 0; k < windowSize[o]; ++k) {
                    if (j + k >= 0 && j + k < nbPositions) {
                        LSExpression classAtPosition = initials[sequence[j + k]];
                        nbCarsWindows[o][j].addOperand(model.at(options, classAtPosition, o));
                    }
                }
            }
        }

        // Number of violations of option o capacity in each window
        objectiveHighPriority = model.sum();
        if (hasLowPriorityOptions)
            objectiveLowPriority = model.sum();
        for (int o = 0; o < nbOptions; ++o) {
            LSExpression nbViolationsWindows = model.sum();
            for (int j = startPosition - windowSize[o] + 1; j < nbPositions; ++j) {
                nbViolationsWindows.addOperand(model.max(0, nbCarsWindows[o][j] - maxCarsPerWindow[o]));
            }
            if (isPriorityOption[o])
                objectiveHighPriority.addOperand(nbViolationsWindows);
            else
                objectiveLowPriority.addOperand(nbViolationsWindows);
        }

        // Color change between position p and position p + 1
        vector<LSExpression> colorChange;
        colorChange.resize(nbPositions - 1);
        objectiveColor = model.sum();
        for (int p = startPosition - 1; p < nbPositions - 1; ++p) {
            LSExpression currentClass = initials[sequence[p]];
            LSExpression nextClass = initials[sequence[p + 1]];
            colorChange[p] = colors[currentClass] != colors[nextClass];
            objectiveColor.addOperand(colorChange[p]);
        }

        // Paint limit constraints: at least one change every paintBatchLimit positions
        for (int p = startPosition; p < nbPositions - paintBatchLimit - 1; ++p) {
            LSExpression nodeOr = model.or_();
            for (int p2 = 0; p2 < paintBatchLimit; ++p2)
                nodeOr.addOperand(colorChange[p + p2]);
            model.addConstraint(nodeOr);
        }

        // Declare the objectives in the correct order
        switch (objectiveOrder) {
        case COLOR_HIGH_LOW:
            model.minimize(objectiveColor);
            model.minimize(objectiveHighPriority);
            model.minimize(objectiveLowPriority);
            break;
        case HIGH_COLOR_LOW:
            model.minimize(objectiveHighPriority);
            model.minimize(objectiveColor);
            model.minimize(objectiveLowPriority);
            break;
        case HIGH_LOW_COLOR:
            model.minimize(objectiveHighPriority);
            model.minimize(objectiveLowPriority);
            model.minimize(objectiveColor);
            break;
        case COLOR_HIGH:
            model.minimize(objectiveColor);
            model.minimize(objectiveHighPriority);
            break;
        case HIGH_COLOR:
            model.minimize(objectiveHighPriority);
            model.minimize(objectiveColor);
            break;
        }

        model.close();

        // Set the initial solution
        sequence.getCollectionValue().clear();
        for (int p = 0; p < nbPositions; ++p)
            sequence.getCollectionValue().add(p);

        // Parametrize the solver
        localsolver.getParam().setTimeLimit(limit);

        localsolver.solve();
    }

    /* Write the solution in a file with the following format:
     * - 1st line: value of the objectives;
     * - 2nd line: for each position p, index of class at positions p. */
    void writeSolution(const string& fileName) {
        ofstream outfile;
        outfile.exceptions(ofstream::failbit | ofstream::badbit);
        outfile.open(fileName.c_str());

        outfile << objectiveColor.getValue() << " ";
        outfile << objectiveHighPriority.getValue() << " ";
        outfile << objectiveLowPriority.getValue() << endl;
        for (int p = 0; p < nbPositions; ++p) {
            outfile << initialSequence[sequence.getCollectionValue().get(p)] << " ";
        }
        outfile << endl;
    }
};

int main(int argc, char** argv) {
    if (argc < 2) {
        cerr << "Usage: car_sequencing_color inputFile [outputFile] [timeLimit]" << endl;
        return 1;
    }

    const char* instanceFile = argv[1];
    const char* outputFile = argc >= 3 ? argv[2] : NULL;
    const char* strTimeLimit = argc >= 4 ? argv[3] : "60";

    try {
        CarSequencingColor model;
        model.readInstance(instanceFile);
        model.solve(atoi(strTimeLimit));
        if (outputFile != NULL)
            model.writeSolution(outputFile);
        return 0;
    } catch (const exception& e) {
        cerr << "An error occurred: " << e.what() << endl;
        return 1;
    }
}
